from typing import Iterable, Iterator, List, TypeVar, Union

import requests
from lxml import etree

from geost.bro.bro_utils import get_bbox_criteria
from geost.projections import horizontal_reference_transformer

Coordinate = TypeVar("Coordinate", int, float)


class BroApi:
    apis = {
        "CPT": "/sr/cpt/v1",
        "BHR-P": "/sr/bhrp/v2",
        "BHR-GT": "/sr/bhrgt/v2",
        "BHR-G": "/sr/bhrg/v3",
        "SFR": "/sr/sfr/v2",
    }

    def __init__(self, server_url=r"https://publiek.broservices.nl"):
        self.session = requests.Session()
        self.server_url = server_url
        self.search_url = "/characteristics/searches"
        self.object_list = []

    def get_objects(
        self, bro_ids: str | Iterable, object_type: str = "CPT"
    ) -> list[bytes]:
        """
        Retrieve BRO objects via API requests to the public BRO REST-API.

        Parameters
        ----------
        bro_ids : str | Iterable
            Single BRO ID or Iterable of BRO ID's
        object_type : str, optional
            BRO object type. Can be CPT (Cone penetration tests), BHR-P (Soil cores),
            BHR-GT (Geotechnical cores), or BHR-G (Geological cores). By default "CPT".

        Returns
        -------
        list[bytes]
            List of bytestrings containing the XML data of the requested BRO objects.

        Raises
        ------
        Warning
            When a non-existing BRO ID was given.
        Warning
            When the server does not respond to the request (40x error).
        """
        if isinstance(bro_ids, str):
            bro_ids = [bro_ids]

        api_response = []
        for bro_id in bro_ids:
            response = self.session.get(
                f"{self.server_url}{self.apis[object_type]}/objects/{bro_id}"
            )
            if response.status_code == 200 and "rejection" not in response.text:
                api_response.append(response.text.encode("utf-8"))
            elif response.status_code != 200 or "rejection" in response.text:
                raise Warning(
                    f"Error {response.status_code}: Unable to request {bro_id} from ",
                    "database",
                )
        return api_response

    def search_objects_in_bbox(
        self,
        xmin: Coordinate,
        ymin: Coordinate,
        xmax: Coordinate,
        ymax: Coordinate,
        epsg: str = "28992",
        object_type: str = "CPT",
    ) -> List[str]:
        """
        Search for BRO objects of the given object type within the given bounding box.
        Returns a list of BRO objects that can be used to retrieve their data using the
        get_objects method.

        Parameters
        ----------
        xmin : Union[float, int]
            x-coordinate of the bbox lower left corner.
        ymin : Union[float, int]
            y-coordinate of the bbox lower left corner.
        xmax : Union[float, int]
            x-coordinate of the bbox upper right corner.
        ymax : Union[float, int]
            y-coordinate of the bbox upper right corner.
        epsg : str, optional
            Coordinate reference system of the given bbox coordinates, by default
            "28992" (= Rijksdriehoek New CRS).
        object_type : str, optional
            BRO object type. Can be CPT (Cone penetration tests), BHR-P (Soil cores),
            BHR-GT (Geotechnical cores), or BHR-G (Geological cores). By default "CPT".

        Returns
        -------
        List[str]
            List containing BRO ID's of objects found within the bbox.

        Raises
        ------
        Warning
            If the server does not respond or the search query was rejected.
        """
        response = self.__response_to_bbox(
            xmin, ymin, xmax, ymax, epsg=epsg, object_type=object_type
        )
        if response.status_code == 200 and "rejection" not in response.text:
            etree_root = etree.fromstring(response.text.encode("utf-8"))
            self.object_list += self.__objects_from_etree(etree_root)
        elif response.status_code == 400 or "groter dan 2000" in response.text:
            self.__search_objects_in_divided_bbox(
                xmin, ymin, xmax, ymax, epsg=epsg, object_type=object_type
            )
        else:
            raise Warning(
                "Selection is invalid and could not be retrieved from the database"
            )

    def __search_objects_in_divided_bbox(
        self,
        xmin: Coordinate,
        ymin: Coordinate,
        xmax: Coordinate,
        ymax: Coordinate,
        epsg: str = "28992",
        object_type: str = "CPT",
    ) -> List[str]:
        division_levels = int((xmax - xmin + ymax - ymin) / 1000)
        division_x = (xmax - xmin) / division_levels
        for division_level in range(division_levels):
            print(
                f"More than 2000 object requests in API call, dividing calls. Current call {division_level + 1}/{division_levels}"
            )
            xmin_divided = xmin + (division_level * division_x)
            xmax_divided = xmin + ((division_level + 1) * division_x)
            response = self.__response_to_bbox(
                xmin_divided,
                ymin,
                xmax_divided,
                ymax,
                epsg=epsg,
                object_type=object_type,
            )
            etree_root = etree.fromstring(response.text.encode("utf-8"))
            self.object_list += self.__objects_from_etree(etree_root)

    def __response_to_bbox(
        self,
        xmin: Coordinate,
        ymin: Coordinate,
        xmax: Coordinate,
        ymax: Coordinate,
        epsg: str = "28992",
        object_type: str = "CPT",
    ):
        transformer = horizontal_reference_transformer(epsg, 4326)
        xmin, ymin = transformer.transform(xmin, ymin)
        xmax, ymax = transformer.transform(xmax, ymax)
        criteria = get_bbox_criteria(xmin, ymin, xmax, ymax)
        api_url = self.server_url + self.apis[object_type] + self.search_url
        response = self.session.post(api_url, json=criteria)
        return response

    def __objects_from_etree(self, etree_root):
        namespace = etree_root.nsmap["brocom"]
        bro_ids = etree_root.findall(".//{" + f"{namespace}" + "}broId")
        bro_objects = [id.text for id in bro_ids]
        return bro_objects



from pathlib import Path
from typing import Iterable, List, Literal

import rioxarray as rio
import xarray as xr

from geost.enums import UnitEnum
from geost.models import VoxelModel

from .bro_utils import coordinates_to_cellcenters, flip_ycoordinates


class Lithology(UnitEnum):  # pragma: no cover
    anthropogenic = 0
    organic = 1
    clay = 2
    loam = 3
    fine_sand = 5
    medium_sand = 6
    coarse_sand = 7
    gravel = 8
    shells = 9


class AntropogenicUnits(UnitEnum):  # pragma: no cover
    AAOP = 1000
    AAES = 1005


class HoloceneUnits(UnitEnum):  # pragma: no cover
    NIGR = 1010
    NINB = 1045
    NASC = 1020
    ONAWA = 1030
    NAZA = 1040
    NAWAZU = 1048
    NAWAAL = 1049
    NAWA = 1050
    BHEC = 1060
    OEC = 1070
    NAWOBE = 1080
    KK1 = 1085
    NIFL = 1089
    NIHO = 1090
    NAZA2 = 1095
    NAWO = 1100
    NWNZ = 1110
    NAWOVE = 1120
    KK2 = 1125
    NIBA = 1130
    NA = 2000
    EC = 2010
    NI = 2020
    KK = 2030


class OlderUnits(UnitEnum):  # pragma: no cover
    BXKO = 3000
    BXSI = 3010
    BXSI1 = 3011
    BXWI = 3020
    BXSI2 = 3012
    BXWIKO = 3025
    BXWISIKO = 3030
    BXDE = 3040
    BXDEKO = 3045
    BXSC = 3050
    BXLM = 3060
    BXBS = 3090
    BX = 3100
    KRWY = 4000
    KRBXDE = 4010
    KRZU = 4020
    KROE = 4030
    KRTW = 4040
    KR = 4050
    BEOM = 4055
    BEWY = 4060
    BERO = 4070
    BE = 4080
    KW1 = 4085
    KW = 4090
    WB = 4100
    EE = 4110
    EEWB = 4120
    DR = 5000
    DRGI = 5010
    GE = 5020
    DN = 5030
    URTY = 5040
    PE = 5050
    UR = 5060
    ST = 5070
    AP = 5080
    SY = 5090
    PZ = 5100
    WA = 5110
    PZWA = 5120
    MS = 5130
    KI = 5140
    OO = 5150
    IE = 5160
    VI = 5170
    BR = 5180
    VE = 5185
    RUBO = 5190
    RU = 5200
    TOZEWA = 5210
    TOGO = 5220
    TO = 5230
    DOAS = 5240
    DOIE = 5250
    DO = 5260
    LA = 5270
    HT = 5280
    HO = 5290
    MT = 5300
    GU = 5310
    VA = 5320
    AK = 5330


class ChannelBeltUnits(UnitEnum):  # pragma: no cover
    AEC = 6000
    ABEOM = 6005
    ANAWA = 6010
    ANAWO = 6020
    BEC = 6100
    BNAWA = 6110
    BNAWO = 6120
    CEC = 6200
    CNAWA = 6210
    CNAWO = 6220
    DEC = 6300
    DNAWA = 6310
    DNAWO = 6320
    EEC = 6400
    ENAWA = 6410
    ENAWO = 6420


class StratGeotop:
    """
    Class for making selections of stratigraphic units in GeoTop containing class attributes
    four main groups of stratigraphic units:

    - holocene : alias of :class:`~geost.bro.bro_geotop.HoloceneUnits`
    - channel : alias of :class:`~geost.bro.bro_geotop.ChannelBeltUnits`
    - older : alias of :class:`~geost.bro.bro_geotop.OlderUnits`
    - antropogenic : alias of :class:`~geost.bro.bro_geotop.AntropogenicUnits`

    Each of these class attributes is a :class:`~geost.enums.UnitEnum` which contains
    abbreviation (e.g. "NAWA") and corresponding number for geological units.

    Example usage
    -------------
    Select a group of "tidal" units by name:

    >>> tidal = StratGeotop.select_units(["NAWA", "ANAWA", "NAWO"])

    Or select a group of values:

    >>> selection = StratGeotop.select_values([1100, 6010, 6100])

    """

    holocene = HoloceneUnits
    channel = ChannelBeltUnits
    older = OlderUnits
    antropogenic = AntropogenicUnits

    @classmethod
    def select_units(cls, units: str | Iterable[str]):
        """
        Select units by name.

        Parameters
        ----------
        units : str | Iterable[str]
            Name as string or array_like object of strings with units to select.

        Returns
        -------
        List[enum]
            List of the selected units.

        Examples
        --------
        Select units based on name:

        >>> nawa = StratGeotop.select_units("NAWA")

        Or select a group of "tidal" units:

        >>> tidal = StratGeotop.select_units(["NAWA", "ANAWA", "NAWO"])

        """
        units = cls.holocene.cast_to_list(units)
        selection = (
            cls.holocene.select_units(units)
            + cls.channel.select_units(units)
            + cls.older.select_units(units)
            + cls.antropogenic.select_units(units)
        )
        return selection

    @classmethod
    def select_values(cls, values: int | Iterable[int]):
        """
        Select units by value.

        Parameters
        ----------
        units : str | Iterable[str]
            Number or array_like object of numbers with units to select.

        Returns
        -------
        List[enum]
            List of the selected units.

        Examples
        --------
        Select units based on value:

        >>> selection = StratGeotop.select_values(1100)

        Or select a group of values:

        >>> selection = StratGeotop.select_units([1100, 6010, 6100])

        """
        values = cls.holocene.cast_to_list(values)
        selection = (
            cls.holocene.select_values(values)
            + cls.channel.select_values(values)
            + cls.older.select_values(values)
            + cls.antropogenic.select_values(values)
        )
        return selection

    @classmethod
    def to_dict(cls, key: Literal["unit", "value"] = "value") -> dict:
        """
        Create a mapping dictionary from StratGeotop with "unit": "value" or "value": "unit"
        mapping.

        Parameters
        ----------
        key : Literal["unit", "value"], optional
            Determine what to use for "key" ("unit" or "value"). The default is "value".

        Returns
        -------
        dict

        """
        return (
            cls.holocene.to_dict(key)
            | cls.channel.to_dict(key)
            | cls.older.to_dict(key)
            | cls.antropogenic.to_dict(key)
        )


class GeoTop(VoxelModel):
    @classmethod
    def from_netcdf(
        cls,
        nc_path: str | Path,
        data_vars: List[str] = None,
        bbox: tuple[float, float, float, float] = None,
        lazy: bool = True,
        **xr_kwargs,
    ):
        """
        Read the BRO GeoTop subsurface model from a netcdf dataset into a GeoTop
        VoxelModel instance. The complete dataset of GeoTop can be downloaded from:
        https://dinodata.nl/opendap.

        Parameters
        ----------
        nc_path : str | Path
            Path to the netcdf file of GeoTop.
        data_vars : ArrayLike
            List or array-like object specifying which data variables to return.
        bbox : tuple (xmin, ymin, xmax, ymax), optional
            Specify a bounding box (xmin, ymin, xmax, ymax) to return a selected area of
            GeoTop. The default is None.
        lazy : bool, optional
            If True, netcdf loads lazily. Use False for speed improvements for larger
            areas but that still fit into memory. The default is False.
        **xr_kwargs
            Additional keyword arguments xarray.open_dataset. See relevant documentation
            for details.

        Returns
        -------
        GeoTop
            GeoTop instance of the netcdf file.

        """
        cellsize = 100
        dz = 0.5
        crs = 28992

        if lazy and "chunks" not in xr_kwargs:
            xr_kwargs["chunks"] = "auto"

        ds = xr.open_dataset(nc_path, **xr_kwargs)
        ds = coordinates_to_cellcenters(ds, cellsize, dz)

        if bbox is not None:
            xmin, ymin, xmax, ymax = bbox
            ds = ds.sel(
                x=slice(xmin - (0.5 * cellsize), xmax + (0.5 * cellsize)),
                y=slice(ymin - (0.5 * cellsize), ymax + (0.5 * cellsize)),
            )

        if data_vars is not None:
            ds = ds[data_vars]

        if not lazy:
            print("Load data")
            ds = ds.load()

        ds = flip_ycoordinates(ds)
        ds = ds.rio.write_crs(crs)

        return cls(ds)

    @classmethod
    def from_opendap(
        cls,
        url: str = r"https://dinodata.nl/opendap/GeoTOP/geotop.nc",
        data_vars: List[str] = None,
        bbox: tuple = None,
        lazy: bool = True,
        **xr_kwargs,
    ):
        """
        Download an area of GeoTop directly from the OPeNDAP data server into a GeoTop
        VoxelModel instance.

        Parameters
        ----------
        url : str
            Url to the netcdf file on the OPeNDAP server. See:
            https://www.dinoloket.nl/modelbestanden-aanvragen
        data_vars : ArrayLike
            List or array-like object specifying which data variables to return.
        bbox : tuple (xmin, ymin, xmax, ymax), optional
            Specify a bounding box (xmin, ymin, xmax, ymax) to return a selected area of
            GeoTop. The default is None but for practical reasons, specifying a bounding
            box is advised (TODO: find max downloadsize for server).
        lazy : bool, optional
            If True, netcdf loads lazily. Use False for speed improvements for larger
            areas but that still fit into memory. The default is False.
        **xr_kwargs
            Additional keyword arguments xarray.open_dataset. See relevant documentation
            for details.

        Returns
        -------
        GeoTop
            GeoTop instance for the selected area.

        """
        return cls.from_netcdf(url, data_vars, bbox, lazy, **xr_kwargs)



from typing import Iterable

import numpy as np
import pandas as pd
import pyvista as pv
import xarray as xr
from pyvista import CellType, MultiBlock


def prepare_borehole(borehole: pd.DataFrame, vertical_factor: float) -> np.ndarray:
    borehole_xyz = borehole[["x", "y", "bottom"]].to_numpy()
    surface_xyz = np.array(
        [borehole_xyz[0, 0], borehole_xyz[0, 1], borehole["surface"].iloc[0]]
    )
    borehole_xyz = np.vstack([surface_xyz, borehole_xyz], dtype="float64")
    borehole_xyz[:, 2] *= vertical_factor
    return borehole_xyz


def generate_cylinders(
    table: pd.DataFrame,
    data_columns: list[str],
    radius: float,
    vertical_factor: float,
) -> Iterable:
    boreholes = table.groupby("nr")
    for _, borehole in boreholes:
        borehole_prepared = prepare_borehole(borehole, vertical_factor)
        poly = pv.PolyData(borehole_prepared)
        line_segments = np.arange(0, len(borehole_prepared), dtype=np.int_)
        line_segments = np.insert(line_segments, 0, len(borehole_prepared))
        poly.lines = line_segments

        for data_column in data_columns:
            poly[data_column] = np.hstack(
                [borehole[data_column].values[0], borehole[data_column].values]
            )
        cylinder = poly.tube(radius=radius)
        yield cylinder


def borehole_to_multiblock(
    table: pd.DataFrame,
    data_columns: list[str],
    radius: float,
    vertical_factor: float,
) -> MultiBlock:
    """
    Create a PyVista MultiBlock object from the parsed boreholes/cpt's.

    Parameters
    ----------
    table : pd.DataFrame
        Table of borehole/CPT objects. This is CptCollection.data or
        BoreholeCollection.data.
    data_columns : List[str]
        Column names of data arrays to write in the vtk file
    radius : float
        Radius of borehole cylinders
    vertical_factor : float
        Vertical adjustment factor to convert e.g. heights in cm to m.

    Returns
    -------
    pv.MultiBlock
        MultiBlock object with boreholes represented as cylinder geometries

    """
    cylinders = generate_cylinders(table, data_columns, radius, vertical_factor)
    cylinders_multiblock = pv.MultiBlock(list(cylinders))
    return cylinders_multiblock


def layerdata_to_pyvista_unstructured(
    layerdata: pd.DataFrame,
    displayed_variables: list[str],
    radius: float = 1.0,
) -> pv.UnstructuredGrid:
    """
    Convert a layerdata object to a PyVista UnstructuredGrid.

    Parameters
    ----------
    layerdata : LayerData
        The input layerdata object containing 3D grid data with coordinates 'x', 'y',
        and 'z'.

    Returns
    -------
    grid : pyvista.UnstructuredGrid
        The resulting voxel model as a PyVista UnstructuredGrid.
    """
    # Get the data from the layerdata object
    x = layerdata["x"].values
    y = layerdata["y"].values
    top = layerdata["surface"].values - layerdata["top"].values
    bottom = layerdata["surface"].values - layerdata["bottom"].values

    # Define all cell corner coordinates in the required order
    voxels = np.array(
        [
            [x - radius, y - radius, bottom],
            [x + radius, y - radius, bottom],
            [x - radius, y + radius, bottom],
            [x + radius, y + radius, bottom],
            [x - radius, y - radius, top],
            [x + radius, y - radius, top],
            [x - radius, y + radius, top],
            [x + radius, y + radius, top],
        ]
    )
    voxels = np.rollaxis(voxels, -1, 0)
    points = voxels.reshape(-1, 3)

    # Create the cells array
    n_voxels = len(x)
    cells_voxel = np.arange(n_voxels * 8).reshape((n_voxels, 8))

    # Create unstructured grid and assign data variables
    grid = pv.UnstructuredGrid({CellType.VOXEL: cells_voxel}, points)
    for var in displayed_variables:
        if var not in layerdata.columns:
            print(
                f"Variable '{var}' is unavailable in the layerdata. Skipping this variable."
            )
            continue
        data = layerdata[var].values
        grid.cell_data[var] = data.flatten(order="C")

    return grid


def voxelmodel_to_pyvista_structured(
    dataset: xr.Dataset,
    resolution: tuple[float, float, float],
    displayed_variables: list[str] = None,
) -> pv.StructuredGrid:
    """
    Convert an xarray dataset to a PyVista voxel model (StructuredGrid).

    Parameters
    ----------
    dataset : xarray.Dataset
        The input dataset containing 3D grid data with coordinates 'x', 'y', and 'z'.
    displayed_variables : list of str
        List of variable names in the dataset to include as cell data in the voxel model.
    resolution : tuple of float
        The resolution (dy, dx, dz) representing the size of each voxel along the y, x,
        and z axes.

    Returns
    -------
    grid : pyvista.StructuredGrid
        The resulting voxel model as a PyVista StructuredGrid, with cell data assigned
        for each variable in `displayed_variables`.
    """
    if displayed_variables is None:
        displayed_variables = dataset.data_vars

    # Check if the dataset has the required dimensions and order. Because we treat the
    # voxelmodel as image data, we need dimensions in the order (x, y, z) with
    # coordinates in increasing order with respect to the origin.
    dataset = check_voxelmodel_dims(dataset, dim_order=("x", "y", "z"))

    # Define PyVista structured grid using ImageData
    grid = pv.ImageData()
    grid.spacing = resolution[1], resolution[0], resolution[2]
    grid.origin = dataset.x.values[0], dataset.y.values[0], dataset.z.values[0]
    grid.dimensions = (
        np.array(
            [dataset.sizes["x"], dataset.sizes["y"], dataset.sizes["z"]], dtype=int
        )
        + 1
    )

    # Add the variables to the grid
    for var in displayed_variables:
        if not all(dim in dataset[var].sizes for dim in ["x", "y", "z"]):
            print(
                f"Variable '{var}' does not have the required dimensions 'x', 'y', and "
                "'z'. Skipping this variable."
            )
            continue
        data = dataset[var].values
        grid.cell_data[var] = data.flatten(order="F")

    return grid


def voxelmodel_to_pyvista_unstructured(
    dataset: xr.Dataset,
    resolution: tuple[float, float, float],
    displayed_variables: list[str] = None,
) -> pv.UnstructuredGrid:
    """
    Convert an xarray dataset to a PyVista voxel model (UnstructuredGrid).

    Parameters
    ----------
    dataset : xarray.Dataset
        The input dataset containing 3D grid data with coordinates 'x', 'y', and 'z'.
    displayed_variables : list of str
        List of variable names in the dataset to include as cell data in the voxel model.
    resolution : tuple of float
        The resolution (dx, dy, dz) representing the size of each voxel along the x, y,
        and z axes.

    Returns
    -------
    grid : pyvista.UnstructuredGrid
        The resulting voxel model as a PyVista UnstructuredGrid, with cell data assigned
        for each variable in `displayed_variables`.
    """
    if displayed_variables is None:
        displayed_variables = dataset.data_vars

    # Check if the dataset has the required dimensions and order
    dataset = check_voxelmodel_dims(dataset, dim_order=("y", "x", "z"))

    # Get half resolution (used to compute voxel corners with respect to cell centers)
    xres, yres, zres = resolution[0] / 2, resolution[1] / 2, resolution[2] / 2

    # Compute cell centers (Making use of the fact that xarray coordinates represent
    # those cell centers, so we can use them directly with meshgrid)
    xx, yy, zz = np.meshgrid(dataset.x.values, dataset.y.values, dataset.z.values)
    cell_centers = np.stack([xx, yy, zz], axis=-1).reshape(-1, 3)

    # Compute voxel corners by offsetting cell centers with half resolution in all
    # directions.
    offsets = np.array(
        [
            [-xres, -yres, -zres],
            [xres, -yres, -zres],
            [-xres, yres, -zres],
            [xres, yres, -zres],
            [-xres, -yres, zres],
            [xres, -yres, zres],
            [-xres, yres, zres],
            [xres, yres, zres],
        ]
    )
    voxels = cell_centers[:, None, :] + offsets[None, :, :]
    points = voxels.reshape(-1, 3)

    # Build cell connectivity
    n_voxels = cell_centers.shape[0]
    cells_voxel = np.arange(n_voxels * 8).reshape((n_voxels, 8))

    # Create unstructured grid and assign data variables
    grid = pv.UnstructuredGrid({CellType.VOXEL: cells_voxel}, points)
    for var in displayed_variables:
        if not all(dim in dataset[var].dims for dim in ["x", "y", "z"]):
            print(
                f"Variable '{var}' does not have the required dimensions 'x', 'y', and "
                "'z'. Skipping this variable."
            )
            continue
        data = dataset[var].values
        grid.cell_data[var] = data.flatten(order="C")

    # Take the first data var to check for NaN values and extract only cells in grid
    # that are not NaN. (perhaps this should be done for all data vars?)
    nan_mask = np.isnan(grid.cell_data[grid.cell_data.keys()[0]])
    if np.any(nan_mask):
        grid = grid.extract_cells(~nan_mask)

    return grid


def check_voxelmodel_dims(
    dataset: xr.Dataset, dim_order: tuple = ("x", "y", "z")
) -> xr.Dataset:
    """
    Check if the voxelmodel dataset dimensions are in the required order and contain
    the required dimensions 'x', 'y', and 'z'.

    - If dimensions are missing or wrongly named, raise a ValueError.
    - If the dimensions are in the wrong order, reorder them to dim_order.
    - If the coordinates are not in increasing order, sort them.

    Unless a ValueError was raised, this will return the dataset with the correct
    dimensions. If the dataset is already in the correct order, it will be returned
    unchanged.

    Parameters
    ----------
    dataset : xarray.Dataset
        The input dataset to check.
    dim_order : tuple of str, optional
        The required order of dimensions. Default is ('x', 'y', 'z').

    Returns
    -------
    xr.Dataset
        The dataset with the correct dimensions and order.

    Raises
    ------
    ValueError
        If the dataset does not contain the required dimensions 'x', 'y', and 'z'.
    """
    for dim in dim_order:
        if dim not in dataset.dims or dim not in dataset.coords:
            raise ValueError(
                (
                    f"Dataset must contain '{dim}' dimension. Make sure that this "
                    "spatial dimension exists in the dataset or if it has a different "
                    "name use xarray.Dataset.rename() to rename the corresponding "
                    f"dimension to '{dim}'."
                )
            )

    # Order dataset dimensions to match required order.
    if tuple(dataset.sizes.keys()) != dim_order:
        dataset = dataset.transpose(*dim_order)
    # Also transpose all data_vars to match the required order. You have to do this
    # because the above dataset.transpose() call will not transpose the data_vars.
    for var in dataset.data_vars:
        if tuple(dataset[var].sizes) != dim_order and all(
            [dim in dataset[var].sizes for dim in dim_order]
        ):
            dataset[var] = dataset[var].transpose(*dim_order)

    # Ensure all coordinates are increasing and sort the data if needed.
    for dim in dim_order:
        coord = dataset.coords[dim]
        if coord[0] > coord[-1]:
            dataset = dataset.sortby(dim)

    return dataset



from geost.io.xml import resolvers

BRO = {
    "payload_root": "dispatchDocument",
    "nr": {"xpath": "brocom:broId"},
    "location": {
        "xpath": "deliveredLocation/bhrgtcom:location/gml:Point/gml:pos",
        "resolver": resolvers.parse_coordinates,
        "el-attr": "text",
    },
    "crs": {
        "xpath": "deliveredLocation/bhrgtcom:location/gml:Point",
        "resolver": resolvers.parse_crs,
    },
    "surface": {
        "xpath": "deliveredVerticalPosition/bhrgtcom:offset",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "vertical_datum": {
        "xpath": "deliveredVerticalPosition/bhrgtcom:verticalDatum",
        "el-attr": "text",
    },
    "groundwater_level": {
        "xpath": "boring/bhrgtcom:groundwaterLevel",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "end": {
        "xpath": "boring/bhrgtcom:finalDepthBoring",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "data": {
        "xpath": "boreholeSampleDescription/bhrgtcom:descriptiveBoreholeLog",
        "resolver": resolvers.process_bhrgt_data,
        "layer-attributes": ["upperBoundary", "lowerBoundary", "geotechnicalSoilName"],
    },
}

WIERTSEMA = {
    "payload_root": "ns1:sourceDocument",
    "nr": {"xpath": "ns1:objectIdAccountableParty"},
    "location": {
        "xpath": "ns1:deliveredLocation/ns3:location/ns2:Point/ns2:pos",
        "resolver": resolvers.parse_coordinates,
        "el-attr": "text",
    },
    "crs": {
        "xpath": "ns1:deliveredLocation/ns3:location/ns2:Point",
        "resolver": resolvers.parse_crs,
    },
    "surface": {
        "xpath": "ns1:deliveredVerticalPosition/ns3:offset",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "vertical_datum": {
        "xpath": "ns1:deliveredVerticalPosition/ns3:verticalDatum",
        "el-attr": "text",
    },
    "groundwater_level": {
        "xpath": "ns1:boring/ns3:groundwaterLevel",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "end": {
        "xpath": "ns1:boring/ns3:finalDepthBoring",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "data": {
        "xpath": "ns1:boreholeSampleDescription/ns3:descriptiveBoreholeLog",
        "resolver": resolvers.process_bhrgt_data,
        "layer-attributes": ["upperBoundary", "lowerBoundary", "geotechnicalSoilName"],
    },
}

schemas = {"BRO": BRO, "Wiertsema": WIERTSEMA}



from abc import ABC, abstractmethod


class AbstractHeader(ABC):  # pragma: no cover
    @property
    @abstractmethod
    def gdf(self):
        pass

    @property
    @abstractmethod
    def horizontal_reference(self):
        pass

    @property
    @abstractmethod
    def vertical_reference(self):
        pass

    @gdf.setter
    @abstractmethod
    def gdf(self, gdf):
        pass

    @abstractmethod
    def change_horizontal_reference(self):
        pass

    @abstractmethod
    def change_vertical_reference(self):
        pass

    @abstractmethod
    def get(self):
        pass

    @abstractmethod
    def select_within_bbox(self):
        pass

    @abstractmethod
    def select_with_points(self):
        pass

    @abstractmethod
    def select_with_lines(self):
        pass

    @abstractmethod
    def select_within_polygons(self):
        pass

    @abstractmethod
    def select_by_depth(self):
        pass

    @abstractmethod
    def select_by_length(self):
        pass

    @abstractmethod
    def get_area_labels(self):
        pass


class AbstractData(ABC):  # pragma: no cover
    @property
    @abstractmethod
    def df(self):
        pass

    @property
    @abstractmethod
    def datatype(self):
        pass

    @df.setter
    @abstractmethod
    def df(self, df):
        pass

    @abstractmethod
    def to_header(self):
        pass

    @abstractmethod
    def to_collection(self):
        pass

    @abstractmethod
    def select_by_values(self):
        pass

    @abstractmethod
    def slice_depth_interval(self):
        pass

    @abstractmethod
    def slice_by_values(self):
        pass

    @abstractmethod
    def select_by_condition(self):
        pass

    @abstractmethod
    def get_cumulative_thickness(self):
        # Not sure if this should be here, potentially unsuitable with DiscreteData
        pass

    @abstractmethod
    def get_layer_top(self):
        pass

    @abstractmethod
    def to_datafusiontools(self):
        pass


class AbstractCollection(ABC):  # pragma: no cover
    @property
    @abstractmethod
    def header(self):
        pass

    @property
    @abstractmethod
    def data(self):
        pass

    @property
    def n_points(self):
        pass

    @property
    @abstractmethod
    def horizontal_reference(self):  # Move to header class in future refactor
        pass

    @property
    @abstractmethod
    def vertical_reference(self):  # move to data class in future refactor
        pass

    @header.setter
    @abstractmethod
    def header(self, header):
        pass

    @data.setter
    @abstractmethod
    def data(self, data):
        pass

    @horizontal_reference.setter
    @abstractmethod
    def horizontal_reference(self, to_epsg: int):
        pass

    @vertical_reference.setter
    @abstractmethod
    def vertical_reference(self, to_epsg: str):  # will use epsg after refactor
        pass

    @abstractmethod
    def get(self):
        pass

    @abstractmethod
    def reset_header(self):
        pass

    @abstractmethod
    def check_header_to_data_alignment(self):
        pass

    @abstractmethod
    def select_within_bbox(self):
        pass

    @abstractmethod
    def select_with_points(self):
        pass

    @abstractmethod
    def select_with_lines(self):
        pass

    @abstractmethod
    def select_within_polygons(self):
        pass

    @abstractmethod
    def select_by_depth(self):
        pass

    @abstractmethod
    def select_by_length(self):
        pass

    @abstractmethod
    def select_by_condition(self):
        pass

    @abstractmethod
    def get_area_labels(self):
        pass

    @abstractmethod
    def select_by_values(self):
        pass

    @abstractmethod
    def slice_depth_interval(self):
        pass

    @abstractmethod
    def slice_by_values(self):
        pass

    # @abstractmethod
    # def get_cumulative_thickness(self):
    #     # Not sure if this should be here, potentially unsuitable with DiscreteData
    #     # These kind of methods should go to a seperate layer_analysis module with
    #     # functions to cover such analyses
    #     pass

    # @abstractmethod
    # def get_layer_top(self):
    #     # These kind of methods should go to a seperate layer_analysis module with
    #     # functions to cover such analyses
    #     pass

    @abstractmethod
    def to_pyvista_cylinders(self):
        pass

    @abstractmethod
    def to_pyvista_grid(self):
        pass

    @abstractmethod
    def to_datafusiontools(self):
        pass



import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

import geopandas as gpd
import numpy as np
import rioxarray as rio
import xarray as xr

from geost.export import vtk
from geost.utils import check_geometry_instance

from .model_utils import sample_along_line, sample_with_coords


class AbstractSpatial(ABC):  # pragma: no cover
    @property
    @abstractmethod
    def xmin(self):
        pass

    @property
    @abstractmethod
    def xmax(self):
        pass

    @property
    @abstractmethod
    def ymin(self):
        pass

    @property
    @abstractmethod
    def ymax(self):
        pass

    @property
    @abstractmethod
    def resolution(self):
        pass

    @property
    @abstractmethod
    def crs(self):
        pass

    @abstractmethod
    def sel(self):
        pass

    @abstractmethod
    def isel(self):
        pass

    @abstractmethod
    def select_with_line(self):
        pass

    @abstractmethod
    def select_with_points(self):
        pass

    @abstractmethod
    def select_within_polygons(self):
        pass

    @abstractmethod
    def select_within_bbox(self):
        pass

    @abstractmethod
    def select_by_values(self):
        pass


class AbstractModel3D(ABC):  # pragma: no cover
    @property
    @abstractmethod
    def zmin(self):
        pass

    @property
    @abstractmethod
    def zmax(self):
        pass

    @abstractmethod
    def select_top(self):
        pass

    @abstractmethod
    def select_bottom(self):
        pass

    @abstractmethod
    def slice_depth_interval(self):
        pass

    @abstractmethod
    def select_surface_level(self):
        pass

    @abstractmethod
    def zslice_to_tiff(self):
        pass

    @abstractmethod
    def get_thickness(self):
        pass

    @abstractmethod
    def to_pyvista_grid(self):
        pass


class VoxelModel(AbstractSpatial, AbstractModel3D):
    def __init__(self, ds: xr.Dataset):
        self.ds = ds

    def __getitem__(self, item):
        return self.ds[item]

    def __setitem__(self, key, item):
        self.ds[key] = item

    def __repr__(self):
        instance = f"{self.__class__.__name__}"
        layers = self.ds.data_vars
        dimensions = f"Dimensions: {dict(self.ds.sizes)}"
        resolution = f"Resolution (y, x, z): {self.resolution}"
        return f"{instance}\n{layers}\n{dimensions}\n{resolution}"

    @classmethod
    def from_netcdf(
        cls,
        nc_path: str | Path,
        data_vars: list[str] = None,
        bbox: tuple[float, float, float, float] = None,
        lazy: bool = True,
        **xr_kwargs,
    ):
        """
        Read data from a NetCDF file of a voxelmodel data into a VoxelModel instance.

        This assumes the voxelmodel is according to the following conventions:
        - The coordinate dimensions of the voxelmodel are: "x", "y" (horizontal) and "z"
        (depth).
        - The coordinates in the y-dimension are in descending order.

        Parameters
        ----------
        nc_path : str | Path
            Path to the netcdf file of the voxelmodel.
        data_vars : ArrayLike
            List or array-like object specifying which data variables to return.
        bbox : tuple (xmin, ymin, xmax, ymax), optional
            Specify a bounding box (xmin, ymin, xmax, ymax) to return a selected area of
            the voxelmodel. The default is None.
        lazy : bool, optional
            If True, netcdf loads lazily. Use False for speed improvements for larger
            areas but that still fit into memory. The default is False.
        **xr_kwargs
            Additional keyword arguments xarray.open_dataset. See relevant documentation
            for details.

        Returns
        -------
        VoxelModel
            VoxelModel instance of the netcdf file.

        Examples
        --------
        Read all model data from a local NetCDF file:

        >>> VoxelModel.from_netcdf("my_netcdf_file.nc")

        Read specific data variables and the data within a specific area from the NetCDF
        file:

        >>> VoxelModel.from_netcdf(
        ...     "my_netcdf_file.nc",
        ...     data_vars=["my_var"],
        ...     bbox=(1, 1, 3, 3) # (xmin, ymin, xmax, ymax)
        ... )

        Note that this method assumes the y-coordinates are in descending order. For y-
        ascending coordinates change ymin and ymax coordinates:

        >>> VoxelModel.from_netcdf(
        ...     "my_netcdf_file.nc", bbox=(1, 3, 1, 3) # (xmin, ymax, xmax, ymin)
        ... )

        """
        ds = xr.open_dataset(nc_path, **xr_kwargs)

        if bbox is not None:
            xmin, ymin, xmax, ymax = bbox
            ds = ds.sel(x=slice(xmin, xmax), y=slice(ymax, ymin))

        if data_vars is not None:
            ds = ds[data_vars]

        if not lazy:
            print("Load data")
            ds = ds.load()

        return cls(ds)

    @property
    def xmin(self) -> float:
        return self.horizontal_bounds[0]

    @property
    def xmax(self) -> float:
        return self.horizontal_bounds[2]

    @property
    def ymin(self) -> float:
        return self.horizontal_bounds[1]

    @property
    def ymax(self) -> float:
        return self.horizontal_bounds[3]

    @property
    def zmin(self) -> float:
        return self.vertical_bounds[0]

    @property
    def zmax(self) -> float:
        return self.vertical_bounds[1]

    @property
    def horizontal_bounds(self) -> tuple[float, float, float, float]:
        return self.ds.rio.bounds()

    @property
    def vertical_bounds(self) -> tuple[float, float]:
        if not hasattr(self, "_zmin"):
            self._get_internal_zbounds()
        return float(self._zmin), float(self._zmax)

    @property
    def resolution(self) -> tuple[float, float, float]:
        """
        Return a tuple (dy, dx, dz) of the VoxelModel resolution.
        """
        if not hasattr(self, "_dz"):
            self._get_internal_zbounds()
        dx, dy = np.abs(self.ds.rio.resolution())
        return (float(dy), float(dx), float(self._dz))

    @property
    def sizes(self):
        return self.ds.sizes

    @property
    def shape(self):
        return tuple(self.sizes.values())

    @property
    def crs(self):
        return self.ds.rio.crs

    @property
    def variables(self):
        return self.ds.data_vars

    def _get_internal_zbounds(self):
        self._dz = np.abs(np.diff(self["z"])[0])
        self._zmin = np.min(self["z"].values)
        self._zmax = np.max(self["z"].values)
        self._zmin -= 0.5 * self._dz
        self._zmax += 0.5 * self._dz

    def sel(self, **xr_kwargs):
        """
        Use Xarray selection functionality to select indices along specified dimensions.
        This uses the ".sel" method of an Xarray Dataset.

        Parameters
        ----------
        **xr_kwargs
            xr.Dataset.sel keyword arguments. See relevant Xarray documentation.

        Examples
        --------
        Select a specified coordinates or slice coordinates from the VoxelModel instance:

        >>> selected = voxelmodel.sel(x=[1, 2, 3])  # Using keyword arguments
        >>> selected = voxelmodel.sel({"x": [1, 2, 3]})  # Using a dictionary
        >>> selected = voxelmodel.sel(x=slice(1, 4))  # Using a slice

        Using additional options as well. For instance, when the desired coordinates do
        not exactly match the VoxelModel coordinates, select the nearest:

        >>> selected = voxelmodel.sel(x=[1.1, 2.5, 3.3], method="nearest")

        """
        selected = self.ds.sel(**xr_kwargs)
        return self.__class__(selected)

    def isel(self, **xr_kwargs):
        """
        Use Xarray selection functionality to select indices along specified dimensions.
        This uses the ".isel" method of an Xarray Dataset.

        Parameters
        ----------
        **xr_kwargs
            xr.Dataset.isel keyword arguments. See relevant Xarray documentation.

        Examples
        --------
        Select a specified coordinates or slice coordinates from the VoxelModel instance:

        >>> selected = voxelmodel.isel(x=[1, 2, 3])  # Using keyword arguments
        >>> selected = voxelmodel.isel({"x": [1, 2, 3]})  # Using a dictionary
        >>> selected = voxelmodel.isel(x=slice(1, 4))  # Using a slice

        """
        selected = self.ds.isel(**xr_kwargs)
        return self.__class__(selected)

    def select_with_points(self, points: str | Path | gpd.GeoDataFrame) -> xr.Dataset:
        """
        Select voxel columns at the locations of point geometries.

        Parameters
        ----------
        points : str | Path | gpd.GeoDataFrame
            Geodataframe (or file that can be parsed to a geodataframe) to select with.

        Returns
        -------
        xr.Dataset
            Selected Xarray Dataset with dimensions "idx" and "z". The dimension "idx"
            is the index order of the selection points. The "z" dimension is the original
            depth dimension of the VoxelModel instance. The returned Dataset contains all
            original data variables of the VoxelModel instance.

        Examples
        --------
        To sample a VoxelModel with a GeoDataFrame containing point geometries:

        >>> selected = voxelmodel.select_with_points(point_gdf)

        This way, it is easily possible to sample a VoxelModel at point locations using
        GeoST Header or Collection objects by accessing their "gdf" attributes.

        Using a Header object:

        >>> selected = voxelmodel.select_with_points(Header.gdf)

        Using a Collection object:

        >>> selected = voxelmodel.select_with_points(Collection.header.gdf)

        """
        points = check_geometry_instance(points)

        if "x" in points.columns and "y" in points.columns:
            coords = points[["x", "y"]].values
        else:
            x, y = points["geometry"].x, points["geometry"].y
            coords = np.c_[x, y]

        return sample_with_coords(self.ds, coords)

    def select_with_line(self):  # pragma: no cover
        raise NotImplementedError()

    def select_within_polygons(self):  # pragma: no cover
        raise NotImplementedError()

    def select_within_bbox(self):  # pragma: no cover
        raise NotImplementedError()

    def select_by_values(self):  # pragma: no cover
        raise NotImplementedError()

    def select_top(self):  # pragma: no cover
        raise NotImplementedError()

    def select_bottom(self):  # pragma: no cover
        raise NotImplementedError()

    def slice_depth_interval(self):  # pragma: no cover
        raise NotImplementedError()

    def select_surface_level(self):  # pragma: no cover
        raise NotImplementedError()

    def zslice_to_tiff(self):  # pragma: no cover
        raise NotImplementedError()

    def get_thickness(
        self,
        condition: xr.Dataset,
        depth_range: tuple[float, float] = None,
    ) -> xr.DataArray:
        """
        Generate a thickness array of voxels in the VoxelModel that meet one or more conditions. For example,
        determine the thickness of a specific lithology within a stratigraphic unit (see example usage below).

        Parameters
        ----------
        condition : xr.DataArray
            Boolean DataArray containing that evaluate to True for the desired condition.
            For example: `voxelmodel["lith"] == 1`.
        depth_range : tuple[float, float], optional
            Search for the condition with a specified depth range of the `VoxelModel`. This should be a tuple
            containing the minimum and maximum depth values (in this order!). The default
            is None, which means the entire depth range of the VoxelModel will be used.

        Returns
        -------
        xr.DataArray
            A DataArray containing the generated map based on the specified conditions.
            The DataArray will have dimensions "y" and "x". The values in the DataArray
            represent the thickness of the selected data variable at each location.

        Examples
        --------
        Determine the thickness where the lithology in the `VoxelModel` is equal to 1:

        >>> lith_map = voxelmodel.get_thickness(voxelmodel["lithology"] == 1)

        Or generate an array on more conditions:

        >>> thickness = voxelmodel.get_thickness((voxelmodel["lithology"] == 1) & (voxelmodel["strat"] == 1100))

        Search for the condition or conditions within a specific depth window:

        >>> thickness = voxelmodel.get_thickness((voxelmodel["lithology"] == 1) & (voxelmodel["strat"] == 1100), depth_range=(-10, -20))
        """
        if depth_range:
            zmin, zmax = depth_range
            condition = condition.sel(z=slice(zmin, zmax))

        # Calculate thickness: sum non-NaN values along the z-axis and multiply by dz
        thickness = xr.where(condition, self.resolution[-1], np.nan).sum(dim="z")

        return thickness

    def to_pyvista_grid(
        self, data_vars: str | list[str] = None, structured: bool = True
    ):
        """
        Convert the VoxelModel to a PyVista grid.

        Parameters
        ----------
        data_vars : str | list[str], optional
            String representing one data variable or list of data variables to include
            in the PyVista grid. If None, all data variables are included. The default
            is None.
        structured : bool, optional
            If True, convert to a structured grid. If False, convert to an unstructured
            grid. The default is True.

        Returns
        -------
        pyvista.UnstructuredGrid or pyvista.StructuredGrid
            PyVista grid representation of the VoxelModel.
        """
        if data_vars is None:
            data_vars = self.ds.data_vars
        elif isinstance(data_vars, str):
            data_vars = [data_vars]

        if structured:
            return vtk.voxelmodel_to_pyvista_structured(
                self.ds,
                self.resolution,
                displayed_variables=data_vars,
            )
        else:
            return vtk.voxelmodel_to_pyvista_unstructured(
                self.ds,
                self.resolution,
                displayed_variables=data_vars,
            )


class LayerModel(AbstractSpatial, AbstractModel3D):  # pragma: no cover TODO: add to doc
    def __init__(self):
        raise NotImplementedError("No support of LayerModel yet.")

    @property
    def xmin(self):
        raise NotImplementedError()

    @property
    def xmax(self):
        raise NotImplementedError()

    @property
    def ymin(self):
        raise NotImplementedError()

    @property
    def ymax(self):
        raise NotImplementedError()

    @property
    def bounds(self):
        raise NotImplementedError()

    @property
    def resolution(self):
        raise NotImplementedError()

    @property
    def crs(self):
        raise NotImplementedError()

    def sel(self):
        raise NotImplementedError()

    def isel(self):
        raise NotImplementedError()

    def select_with_line(self):
        raise NotImplementedError()

    def select_with_points(self):
        raise NotImplementedError()

    def select_within_polygons(self):
        raise NotImplementedError()

    def select_within_bbox(self):
        raise NotImplementedError()

    def select_by_values(self):
        raise NotImplementedError()

    @property
    def zmin(self):
        raise NotImplementedError()

    @property
    def zmax(self):
        raise NotImplementedError()

    def select_top(self):
        raise NotImplementedError()

    def select_bottom(self):
        raise NotImplementedError()

    def slice_depth_interval(self):
        raise NotImplementedError()

    def select_surface_level(self):
        raise NotImplementedError()

    def zslice_to_tiff(self):
        raise NotImplementedError()
